/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.ginf.delivery.HTTP;

import org.w3c.rdf.*;
import org.xml.sax.InputSource;

import org.ginf.*;
import org.ginf.helpers.*;
import org.ginf.delivery.*;
import org.ginf.schemas.*;

//import org.ginf.schemas.CoreComm;
//import org.ginf.schemas.HTTP;
//import org.ginf.schemas.RDF;
//import org.ginf.schemas.CoreState;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * An HTTP layer implementation.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public class HTTPLayer extends Thread implements Layer /*, SchemaController*/ {

  public final static int DEFAULT_PORT = 8271; // RD(F)
  int PRIORITY = Thread.NORM_PRIORITY; // 3; // priority of client threads

  protected ServerSocket  listen_socket;
  protected ThreadGroup   threadgroup;
  String name;

  Hashtable connPool = new Hashtable(); // URL -> Connection (client or server conn)

  Hashtable schemaProperties = new Hashtable();
  SchemaController controller;

  // server port it listens to
  int port;

  Layer dispatcher;
  //  Hashtable dispTable = new Hashtable();
  RDFFactory factory;

  public boolean isClient() {
    return listen_socket == null;
  }

  boolean fetchedSchema = false;

  // General constructor
  public HTTPLayer(String name, RDFFactory factory) {

    this.name = name;
    this.factory = factory;

    SchemaRegistry reg = factory.registry();
    reg.addLoader(new CoreComm());
    reg.addLoader(new CoreState());
    reg.addLoader(new HTTP());

    // A ThreadGroup object is created to keep track of existing connections.
    threadgroup = new ThreadGroup("HTTP Connections");

    //    controller = this;
  }

  public HTTPLayer(RDFFactory factory) {
    this(null, factory);
  }
  /*
  public boolean fetchSchema(String url, Triple t, Model m) {
    
    if(RDF.type.equals(t.predicate()) && url.equals(t.object().toString()))
      return true;
    else
      return false;
  }
  */  
  public void setSchemaController(SchemaController c) {
    controller = c;
  }
  
  public void setConnectionPriority(int priority) {

    PRIORITY = priority;
  }
  // client and server. MUST ALWAYS BE CALLED BEFORE USAGE
  //  public void setRDFMSFactory(RDFMSFactory factory) {
  //    this.factory = factory;
  //  }
  /*
  Layer getDispatcher(String what) {

    for(Enumeration en = dispTable.keys(); en.hasMoreElements(); ) {

      String key = (String)en.nextElement();
      if( what.startsWith(key) )
	return (Layer)dispTable.get(key);
    }
    return null;
  }
  */
  // server method
  public void connect(String what, Layer to) {

    // URI prefix
    //    dispTable.put(what, to);

    if("dispatch".equals(what))
      dispatcher = to;
    else {
      throw new IllegalArgumentException(toString() + " cannot connect \"" + what + "\"");
    }
  }

  // server method
  public void acceptConnections() throws IOException {
    acceptConnections(DEFAULT_PORT);
  }

  // server method
  public void acceptConnections(int port) throws IOException {

    // An attempt is made to create a ServerSocket object.
    listen_socket = new ServerSocket(port);

    //    System.out.println("Listening on " + listen_socket);

    this.start();
  }

  // This is the body of the server thread.  It listens constantly for 
  // connection requests from clients.  For each connection, a new 
  // Connection object is created to handle communication through the
  // new Socket.  When a new connection is established, it is added to
  // a vector of existing connections, and List object is updated to
  // reflect the new addition.  The SYNCHRONIZED mechanism is used to 
  // lock the vector of connections to ensure against redundant operations.
  // The Vulture class does the same, so the vulture won't be removing dead 
  // connections while new ones are being added.
  public void run() {

    fetchSchema();
    try {
      while(true) {

	Socket client_socket = listen_socket.accept();
	// assign new state
	//	Resource state = RDFUtil.noname();
	// do not send header, server connection, // no preset dispatcher
	Connection c = new Connection(this, dispatcher, null, client_socket, threadgroup, PRIORITY);
	c.initiatedFromOutside = true;
      }
    } catch (IOException e) {
      System.err.println("[HTTP] Exception while listening for connections:" + e);
    }
  }

  /**
   * If the client constructs a model, we had no opportunity to fetch the schemas
   */
  void fetchSchema() {
    
    if(!fetchedSchema) {
      factory.registry().fetch(HTTP.NS);
      fetchedSchema = true;
    }
  }

  /**
   * This method can be invoked by
   * 1) client wishing to send (first time, next time)
   * 2) server wishing to reply
   *
   * Strategy:
   * Client sends: if there is a HTTP.Connection inside, take its state. If not, create a new state.
   *
   * Find the message to send (must be unique, may be of multiple types)
   * 
   */
  public Model dispatch(Model m, Layer from) throws Exception {

    fetchSchema();

    SchemaModel model = factory.createSchemaModel( m );
    //    factory.createRDFMS().serialize( model, new OutputStreamWriter( System.out ));
    
    // this will be a unique message
    Model msgs = model.find(null, RDF.type, CoreComm.Message);
    if(msgs.size() == 0) {
      //      Model test = model.find(null, RDFS.subClassOf, CoreComm.Message);
      //      factory.createRDFMS().serialize( test, new OutputStreamWriter( System.err ));
      String excMsg = "";
      try {
	model.validate();
      } catch (Exception invalid) {
	excMsg = invalid.getMessage() + "\n";
      }
      throw new Exception(excMsg + toString() + " cannot deliver, no instance of " + CoreComm.Message + " found in the model");
    }

    Resource message = RDFUtil.get1(msgs).subject();
    
    if(message == null)
      throw new RuntimeException(toString() + " to message to send");
    
    URL url = null;
    
    // find state and Connection
    Resource state = null;
    Connection con = null;
    //    try {
    
    // reply via a known state?
    Model states = model.find( message, CoreState.state, null );
    boolean noState = true;
    
    for(Enumeration en = states.elements(); en.hasMoreElements();) {
      Triple t = (Triple)en.nextElement();
      noState = false;
      Resource t_State = (Resource)t.object();
      if(connPool.containsKey(t_State)) {
	state = t_State;
	break;
      }
    }

    if(state == null && !noState) {
      // not known state, but there are another user states
      
      Triple t_stateInfo = model.find1( null, RDF.type, HTTP.Connection );
      
      if(t_stateInfo != null) {
	
	// user wants to create a new state!
	state = t_stateInfo.subject();
      }
    }

    //      System.err.println("HTTP CONN: " + state);
    //      RDFUtil.printTriples( model, System.err );
    
    // do we already have the Connection?
    if(state != null) {
      con = (Connection)connPool.get( state );
    }

    // discard state
    if(RDFUtil.instanceOf(message, CoreState.DiscardState, model)) {
      if(state == null && noState)
	throw new RuntimeException(toString() + " no state given, cannot discard");
      if(con == null)
	throw new RuntimeException(toString() + " unknown state, cannot discard");
      //	System.err.println("HTTP CLOSE");
      connPool.remove(con);
      con.close();

    } else {
      // message to send
      
      // if no Connection found:
      /* Find destination: first look for HTTP.to
       * If not found, look for a CoreState.to starting with http://
       */
      if(con == null) {
	
	String unknownURISchema = null;
	Triple t_to = model.find1( message, HTTP.to, null );
	if(t_to != null) {
	  
	  url = new URL(t_to.object().toString());
	  
	} else {
	  
	  Model destinations = model.find( message, CoreComm.to, null );
	  
	  for(Enumeration en = destinations.elements(); en.hasMoreElements();) {
	    Triple t = (Triple)en.nextElement();
	    if(t.object().toString().startsWith("http://")) {
	      url = new URL(t.object().toString());
	      break;
	    } else
	      unknownURISchema = t.object().toString();
	  }
	}
	// if url is null, don't know what to do...
	if(url == null)
	  if(unknownURISchema == null)
	    throw new RuntimeException(toString() + " unknown state, cannot discard");
	  else
	    throw new RuntimeException(toString() + " do not know how to send to " + unknownURISchema);


	// if state is still null, create a new one
	// User does not want the state to have any particular ID
	//	if(state == null)
	//	  state = RDFUtil.noname();
	// establish a new connection
	// write header!

	// if state is still null, leave it null. We'll get a reply anyway...
	//	System.err.println("---PORT: " + url.getPort());
	con = new Connection( this, from, state,
			      new Socket( url.getHost(), url.getPort() > 0 ? url.getPort() : DEFAULT_PORT ),
			      threadgroup, PRIORITY );
	if(state != null) {
	  //	  System.err.println("---REGISTERING FOREIGN STATE (CLIENT INITIATED): " + state);
	  connPool.put(state, con);
	}
      }
      // send it now
      writeToConnection(model, con);
      
    } // discard or send message

      //    } catch (Exception any) {
      //      any.printStackTrace( System.err );
      //      throw new Exception("[HTTP] Invalid, missing or multiple destination: " + any);
      //    }

    return null;
  }

  void writeToConnection(Model model, Connection con) throws Exception {

    RDFMS p = factory.createRDFMS();
    
    //    p.serialize( model, new OutputStreamWriter( System.out ));
    
    // send header?
    PrintStream ps = new PrintStream( con.out );
    // PrintWriter ps = new PrintWriter( con.out );

    if(!con.initiatedFromOutside) { // request
      //	  System.err.println("POSTING " + url.getFile());
      //	  RDFUtil.printTriples( model, System.err );
      ps.println("POST / HTTP/1.0");
      //	  ps.println("POST " + url.getFile() + " HTTP/1.0");
      ps.print("\nRDF=");
      ps.flush();
      p.serialize( model, new OutputStreamWriter( new HTTPEncoder(con.out) ));
      // terminator
      con.out.write('\n');
      con.out.flush();

    } else { // response

      if(con.replyCount++ == 0) {

	ps.println("HTTP/1.0 200 OK");
	//	ps.println("Connection: keep-alive");
	ps.println("Content-type: multipart/mixed; boundary=" + BOUNDARY + "\n");
	ps.println("--" + BOUNDARY);
      }
      ps.println("Content-type: text/plain\n");
      ps.flush();
      //	    con.firstReply = false;
      //	  }
      p.serialize( model, new OutputStreamWriter( con.out ));
      ps = new PrintStream(con.out);
      //      ps = new PrintWriter(con.out);
      ps.println("--" + BOUNDARY /*+ "--"*/);
      ps.flush();
    }
    // write separator
    //    for(int i=0; i < 4; i++)
    //      con.out.write('\n');
    //	System.err.println("MODEL SENT!");
  }

  static String BOUNDARY = "G_I_N_F-R_D_F/X_M_L-" + System.currentTimeMillis();

  void sendError(String text, Connection con) {
    
    //    System.err.println("SENDING ERROR!!!");
    try {
      Model m = factory.createModel();
      Resource e = RDFUtil.noname();
      m.add( e, RDF.type, CoreComm.Error );
      m.add( e, CoreComm.errorMessage, text );
      //      m.add( e, CoreState.state, state );
      //      m.add( state, RDF.type, HTTP.Connection );
      writeToConnection(m, con);
      //      dispatch( m, this );
    } catch (Exception any) {}
  }

  public String toString() {
    return name != null ? name : "[HTTP Layer]";
  }
}
