/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.ginf.helpers;

import org.ginf.*;
import org.w3c.rdf.*;

import java.util.*;
import java.io.*;

/**
 * A default implementation of the Model interface.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public class ModelImpl implements Model {

  Hashtable triples; // real triples
  Hashtable subjects, predicates, objects; // search object -> Hashtable of Triples
  String uri;
  //  int noname = 0;

  private ModelImpl(String uri, Hashtable triples, Hashtable subjects, Hashtable predicates, Hashtable objects) {
    this.uri = uri;
    this.triples = triples;
    this.subjects = subjects;
    this.predicates = predicates;
    this.objects = objects;
  }

  private ModelImpl(String uri, Enumeration triples) {
    this(uri);
    while( triples.hasMoreElements() ) 
      add( (Triple)triples.nextElement() );
  }

  public ModelImpl(String uri) {

    this();
    this.uri = uri;
  }

  public ModelImpl() {
    triples = new Hashtable();
    subjects = new Hashtable();
    objects = new Hashtable();
    predicates = new Hashtable();
  }

  public void setURI(String uri) {
    this.uri = uri;
  }

  public String getURI() {
    return uri;
  }


  public Object clone() {

    return duplicate();
  }

  Hashtable cloneTable(Hashtable t) {

    Hashtable c = (Hashtable)t.clone();
    // deep cloning
    for(Enumeration en = t.keys(); en.hasMoreElements();) {
      
      Object key = en.nextElement();
      Object obj = t.get(key);
      if(obj instanceof Hashtable)
	// replace with cloned table
	c.put(key, ((Hashtable)obj).clone());
    }
    return c;
  }

  public Model duplicate() {

    return new ModelImpl(uri, (Hashtable)triples.clone(),
			 cloneTable(subjects),
			 cloneTable(predicates),
			 cloneTable(objects));
  }

  public Model create() {

    return new ModelImpl();
  }

  public Enumeration elements() {

    return triples.elements();
  }

  public Triple find1( Resource subject, Resource predicate, RDFnode object ) {

    if(subject != null && predicate != null && object != null) {
      // just a single triple
      return (Triple)triples.get( new Triple( predicate, subject, object ) );
    }

    Model m = find( subject, predicate, object );

    if(m.size() == 1)
      return (Triple)m.elements().nextElement();
    else
      if(m.size() > 1)
	throw new RuntimeException("Model: multiple triples encountered in find1");
    return null;
  }

  /**
   * General matching method.
   * If null values submitted - ANY
   */
  public Model find( Resource subject, Resource predicate, RDFnode object ) {

    if(subject == null && predicate == null && object == null)
      // all of them
      return duplicate();
    
    if(subject != null && predicate != null && object != null) {
      // just a single triple
      Model res = new ModelImpl(uri);
      Triple t = (Triple)triples.get( new Triple( predicate, subject, object ) );
      if(t != null)
	res.add( t );
      return res;
    }

    //    Hashtable shortest = triples;
    Object shortest = triples;
    shortest = getShortest(shortest, subjects, subject);
    shortest = getShortest(shortest, predicates, predicate);
    shortest = getShortest(shortest, objects, object);

    ModelImpl res = new ModelImpl(uri);

    if(shortest == null)
      return res;

    if(shortest instanceof Triple) {
      if(matchTriple( (Triple)shortest, subject, predicate, object ))
	res.add( (Triple)shortest );
      return res;
    }

    // do normal search in the shortest table
    for(Enumeration en = ((Hashtable)shortest).elements(); en.hasMoreElements();) {

      Triple triple = (Triple)en.nextElement();

      if(matchTriple( triple, subject, predicate, object ))
      res.add( triple );
    }
    return res;
  }

  boolean matchTriple(Triple triple, Resource subject, Resource predicate, RDFnode object) {

    if( subject != null && !triple.subject().equals(subject))
      return false;
      
    if(predicate != null && !triple.predicate().equals(predicate))
      return false;
      
    if(object != null && !triple.object().equals(object))
      return false;

    return true;
  }

  //  Hashtable getShortest(Hashtable t1, Hashtable tt, RDFnode value) {
  Object getShortest(Object t1, Hashtable tt, RDFnode value) {

    if(value == null)
      return t1;
    
    // the shortest
    if(t1 == null)
      return null;

    if(t1 instanceof Triple)
      return t1;

    if(value != null) {
      Object obj = tt.get(value);
      if(obj == null)
	return null;
      if(obj instanceof Triple)
	return obj;
      if( ((Hashtable)obj).size() < ((Hashtable)t1).size())
	return obj;
    }
    return t1;

    /*
    if(value != null) {
      Hashtable t = (Hashtable)tt.get(value);
      if(t != null && t.size() < t1.size())
	return t;
    }
    return t1;
    */
  }

  void insColl( Hashtable coll, RDFnode object, Triple t ) {
    
    Object obj = coll.get( object );
    Hashtable tbl = null;
    if(obj == null)
      coll.put( object, t );
    else {
      if(obj instanceof Hashtable)
	tbl = (Hashtable)obj;
      else { // Triple
	coll.put( object, tbl = new Hashtable() );
	tbl.put( obj, obj ); // old value
      }
      tbl.put( t, t );
    }
  
    /*
    Hashtable tbl = (Hashtable)coll.get( object );
    if(tbl == null)
      coll.put( object, tbl = new Hashtable() );
    tbl.put( t, t );
    */
    
    //    System.err.println("Inserted INTO " + hash + ": " + t + "=" + tbl.size());
  }

  public void add( Triple t ) {

    if(contains(t))
      return;

    triples.put(t, t);

    insColl( subjects, t.subject(), t );
    insColl( predicates, t.predicate(), t );
    insColl( objects, t.object(), t );
  }

  public void add(Resource subject, Resource predicate, String object) {

    add( new Triple(predicate, subject, new Literal(object)) );
  }

  public void add(Resource subject, Resource predicate, RDFnode object) {

    add( new Triple(predicate, subject, object) );
  }

  public int size() {

    return triples.size();
  }

  public void remove(Triple t) {

    if(t == null)
      return;

    triples.remove(t);

    delColl( subjects, t.subject(), t );
    delColl( predicates, t.predicate(), t );
    delColl( objects, t.object(), t );
  }

  void delColl( Hashtable coll, RDFnode value, Triple t ) {

    Object obj = coll.get(value);
    if(obj == null)
      return;
    if(obj instanceof Triple)
      coll.remove(value);
    else
      ((Hashtable)obj).remove(t);
  }

  public boolean contains(Triple t) {

    return triples.get(t) != null;
  }

  public Model union(Model m) {

    for(Enumeration en = m.elements(); en.hasMoreElements();)
      add( (Triple)en.nextElement() );
    return this;
  }

  public Model difference(Model m) {

    for(Enumeration en = m.elements(); en.hasMoreElements();)
      remove( (Triple)en.nextElement() );
    return this;
  }

  public String toString() {

    return "RDF model " + (uri != null ? " " + uri : "") + " with " + triples.size() + " triples";
  }

  public static void main(String[] args) {

    Model m1 = new ModelImpl();
    m1.add( new Resource("res1"), new Resource("pred1"), "obj1");
    m1.add( new Resource("res1"), new Resource("pred2"), "obj2");

    Model m2 = new ModelImpl();
    m2.add( new Resource("res2"), new Resource("pred2"), "obj3");
    m2.add( new Resource("res1"), new Resource("pred2"), "obj4");

    System.err.println("Model 1:");
    RDFUtil.printTriples( m1, System.out );

    System.err.println("Model 2:");
    RDFUtil.printTriples( m2, System.out );

    System.err.println("Union:");
    m1.union( m2 );
    RDFUtil.printTriples( m1, System.out );

    System.err.println(" res1, pred2, null : ");
    RDFUtil.printTriples( m1.find( new Resource("res1"), new Resource("pred2"), null ), System.out );

    System.err.println(" null, pred2, null : ");
    RDFUtil.printTriples( m1.find( null, new Resource("pred2"), null ), System.out );

    System.err.println(" res1, pred2, obj2 : ");
    RDFUtil.printTriples( m1.find( new Resource("res1"), new Resource("pred2"), new Literal("obj2") ), System.out );

    System.out.println( "obj1:" + m1.find1( null, new Resource("pred1"), null ) );
    System.out.println( "Triple:" + m1.find1( new Resource("res1"), new Resource("pred2"), new Literal("obj2") ) );
  }

}
