/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.ginf.helpers;

import org.ginf.*;
import org.w3c.rdf.*;
import org.ginf.schemas.RDF;
import org.ginf.schemas.RDFS;

import java.util.*;
import java.io.*;
import java.net.*;

/**
 * Some useful utilities.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public class RDFUtil {

  // implemented in Resource constructor
  //  static long nonameCount = 0;
  static Random rnd = new java.util.Random(System.currentTimeMillis() +
					   new Object().hashCode());
  static String nsPrefix;

  static {
    try {
      InetAddress addr = InetAddress.getLocalHost();
      nsPrefix = "http://" + addr.getHostName() + "/#";
    } catch (Exception any) {
      nsPrefix = "gen:";
    }
  }

  /**
   * Creates a new unique unnamed resource.
   */
  public static Resource noname() {
    return new Resource(nsPrefix + Integer.toHexString( rnd.nextInt() ));
  }
  /*
  public static Resource noname(String namespace) {
    return new Resource(namespace + "#" + Integer.toHexString( rnd.nextInt() ));
  }
  */

  static final int MAX_ORD = 30; // maximal number of cached ordinals
  static Resource[] ords;
  static {
    ords = new Resource[MAX_ORD + 1]; // 0 is wasted
    for(int i=1; i <= MAX_ORD; i++)
      ords[i] = new Resource(RDF.NS + "#_" + i);
  }

  /**
   * Creates an ordinal property (rdf:li, rdf:_N)
   */
  public static Resource ord(int i) {
    
    if(i < 1)
      throw new IllegalArgumentException("Attempt to construct invalid ordinal resource");
    if(i <= MAX_ORD)
      return ords[i];
    else
      return new Resource(RDF.NS + "#_" + i);
  }

  /**
   * Collects ordinal properties into a vector
   */
  public static Vector getContainerProperties(Resource container, Model m) {

    Vector v = new Vector();

    // if container not instance of container, create vector of a single element
    if(!instanceOf(container, RDFS.Container, m)) {
      
      v.addElement(container);

    } else {

      for(int i=1; ;i++) {
	Triple t_o = m.find1( container, ord(i), null );
	if(t_o == null)
	  break;
	//      System.err.println("" + i + ". " + t_o.object());
	v.addElement( t_o.object() );
      }
    }
    return v;
  }

  /**
   * Converts an ordinal property to an integer.
   */
  public static int getOrd(Resource r) {

    if(r == null)
      return -1;

    String uri = r.toString();
    int pos = getNamespaceEnd(uri);
    if(pos > 0 && pos + 2 < uri.length() && isRDF(uri)) {
      try {
	int n = Integer.parseInt(uri.substring(pos+2));
	if(n >= 1)
	  return n;
      } catch (Exception any) {}
    }
    return -1;
  }

  /**
   * Tests if the URI is qualified, i.e. has a namespace prefix.
   */
  public static boolean isQualified(String s) {
    
    int l = s.length()-1;
    do {
      char c = s.charAt(l);
      if(c == '#' || c == ':')
	return true;
      l--;
    } while (l >= 0);
    return false;
  }

  /**
   * Extracts the namespace prefix out of a URI.
   */
  public static String getNS(String uri) {

    int l = getNamespaceEnd(uri);
    return l > 1 ? uri.substring(0, l-1) : null;
  }

  /**
   * Delivers the name out of the URI (without the namespace prefix).
   */
  public static String getName(String uri) {

    return uri.substring(getNamespaceEnd(uri));
  }

  /**
   * Tests if the URI belongs to the RDF syntax/model namespace.
   */
  public static boolean isRDF(String uri) {
    return uri != null && uri.startsWith(RDF.NS);
  }

  /**
   * Tests if the resource belongs to the RDF syntax/model namespace.
   */
  public static boolean isRDF(Resource r) {
    return isRDF(r.toString());
  }

  /**
   * Position of the namespace end
   */
  static int getNamespaceEnd(String uri) {

    int l = uri.length()-1;
    do {
      char c = uri.charAt(l);
      if(c == '#' || c == ':')
	break;
      l--;
    } while (l >= 0);
    l++;
    return l;
  }

  /*
  public static Model subjectInstanceOf(Resource cls, Model model, RDFFactory f) {

    SchemaModel m = f.createSchemaModel( model );
    Model res = f.createModel();
    for(Enumeration en = m.elements(); en.hasMoreElements();) {
      
      Triple t = (Triple)en.nextElement();
      if( m.find1( t.subject(), RDF.type, cls ) != null )
	res.add( t );
    }
    return res;
  }

  public static Model objectInstanceOf(Resource cls, Model model, RDFFactory f) {

    SchemaModel m = f.createSchemaModel( model );
    Model res = f.createModel();
    for(Enumeration en = m.elements(); en.hasMoreElements();) {
      
      Triple t = (Triple)en.nextElement();
      if( m.find1( (Resource)t.object(), RDF.type, cls ) != null )
	res.add( t );
    }
    return res;
  }
  */

  public static boolean instanceOf(Resource r, Resource cls, Model model) {
    return model.find( r, RDF.type, cls ).size() > 0;
  }
  public static boolean subPropertyOf(Resource r, Resource cls, Model model) {
    return r.equals(cls) || model.find( r, RDFS.subPropertyOf, cls ).size() > 0;
  }


  /**
   * Returns the first triple of the model
   */
  public static Triple get1(Model m) {
    if(m == null || m.size() == 0)
      return null;
    //    if(m.size() > 1)
    //      throw new RuntimeException("Model contains more than one triple");
    return (Triple)m.elements().nextElement();
  }

  /**
   * Prints the triples of a model to the given PrintStream.
   */
  public static void printTriples(Model m, PrintStream ps) {

    for(Enumeration en = m.elements(); en.hasMoreElements();) {

      Triple t = (Triple)en.nextElement();
      ps.println ("triple(\""+t.subject()+"\",\""+t.predicate()+"\",\""+t.object()+"\").");
    }
  }

  /**
   * Dumps the model in a serialized form.
   */
  public static void dumpModel(Model m, PrintStream ps, RDFFactory f) throws Exception {
    f.createRDFMS().serialize( m, new OutputStreamWriter(ps) );
  }

  /**
   * Collects the triples of a model in a vector.
   */
  public static Vector getTripleVector(Model m) {

    Vector v = new Vector();
    for(Enumeration en = m.elements(); en.hasMoreElements();) {

      Triple t = (Triple)en.nextElement();
      v.addElement(t);
    }
    return v;
  }

  /**
   * Removes all triples which have something to do with the given namespace
   */
  public static Model removeNamespace( String ns, Model m ) {

    Model res = m.duplicate();
    for(Enumeration en = m.duplicate().elements(); en.hasMoreElements();) {

      Triple t = (Triple)en.nextElement();
      if(t.subject().toString().startsWith( ns ) ||
	 t.predicate().toString().startsWith( ns ) ||
	 t.object().toString().startsWith( ns )) {
	System.err.println("REMOVING TRIPLE: " + t);
	res.remove(t);
      }
    }
    return res;
  }


}
