/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.ginf.helpers;

import java.io.*;

/**
 * This input filter prohibits closing of the
 * original stream. This is important for
 * connections which have to be "kept alive"
 * (An XML/SAX parser may attempt to close the stream
 * after it finishes parsing).
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public class ReadUntilFilter extends FilterInputStream {

  StringBuffer buf = new StringBuffer(); // buffer to unread
  int bufPos = 0;

  String delimiter;
  int pos;

  public ReadUntilFilter(InputStream in) {
    super(in);
  }
  
  public void close() throws IOException {
    // A SAX driver may attempt to close the stream
    // after parsing. We do not allow this!
    //    in.close();
  }

  public void initDelimiter(String str) {
    delimiter = str;
    pos = 0;
  }

  public void initDelimiter() {
    pos = 0;
  }

  int popBuf() {
    // unread one char from buffer
    int c = buf.charAt(bufPos++);
    if(bufPos >= buf.length()) {
      // last read, clean buffer
      bufPos = 0;
      buf.setLength(0);
    }
    return c;
  }

  public int read() throws IOException {
    
    if(delimiter == null)
      // no delimiter
      return in.read();

    // check if we are done
    if(pos >= delimiter.length())
      return -1;

    // buffer is not empty
    if(buf.length() > 0)
      return popBuf();

    int c = in.read();
    if(c == -1)
      return -1;
    
    // necessarily reads one character more than needed
    while(pos < delimiter.length() && (char)c == delimiter.charAt(pos)) {
      buf.append((char)c);
      c = in.read();
      pos++;
    }
    
    if(pos >= delimiter.length()) {
      // we are done, remove delimiter from buffer
      buf.setLength(0);
      return -1;
    } else {
      pos = 0;
      // to speed up a bit
      if(buf.length() == 0)
	return c;
      else { // we read something into the buffer
	buf.append((char)c); // to be able to continue again
	return popBuf();
      }
    }
  }

  /**
   * Tell whether this stream supports the mark() operation.
   */
  public boolean markSupported() {
    return false;
  }

  /**
   * Read characters into a portion of an array, one by one.
   *
   * @exception  IOException  If an I/O error occurs
   */
  public int read(byte cbuf[], int off, int len) throws IOException {
    
    int i = 0;
    while (len > 0) {
      int c = read();
      if(c >= 0)
	cbuf[off + i++] = (byte)c;
      else
	return i > 0 ? i : -1;
      len--;
    }
    return i;
  }
  
}
